import datetime
import json
import os
import typing as T

from stdl import dt
from stdl.fs import File, json_load, yaml_load

from strto.constants import FROM_FILE_PREFIX, ITER_SEP, SLICE_SEP


class Parser(T.Protocol):
    """
    Base class for all parsers.
    """

    def __call__(self, value: str) -> T.Any:
        value = self.clean(value)
        return self.parse(value)

    def clean(self, value) -> str:
        if isinstance(value, str):
            return value.strip()
        return value

    def parse(self, value: str) -> T.Any:
        raise NotImplementedError


class Cast(Parser):
    """Cast a value to a type."""

    def __init__(self, t: type):
        self.t = t

    def parse(self, value) -> T.Any:
        if isinstance(value, self.t):
            return value
        return self.t(value)


class IterableParser(Parser):
    """
    Convert a value to an iterable.

    Args:
        t (type): The type to of the iterable. Defaults to list.
        sep (str): The separator to split the value by.
        from_file (bool): Whether to allow the value to be a readable from a file.
    """

    def __init__(self, t: type = None, sep: str = ITER_SEP, from_file: bool = False):  # type: ignore
        self.t = t
        self.sep = sep
        self.from_file = from_file

    def parse(self, value) -> T.Iterable:
        if isinstance(value, str):
            if self.from_file and value.startswith(FROM_FILE_PREFIX):
                filepath = value[len(FROM_FILE_PREFIX) :]
                if os.path.isfile(filepath):
                    value = self.read_from_file(filepath)
                else:
                    raise FileNotFoundError(value)
            else:
                value = [i.strip() for i in value.split(self.sep)]
        elif isinstance(value, T.Iterable):
            value = [i.split(self.sep) for i in value]
        else:
            raise TypeError(f"Cannot convert '{value}' to an iterable")
        if self.t is not None:
            return self.t(value)  # type: ignore
        return value

    def read_from_file(self, value: str):
        data = File(value).should_exist().splitlines()
        if len(data) == 1 and self.sep in data[0]:
            data = data[0].split(self.sep)
        return [i.strip() for i in data]


class MappingParser(IterableParser):
    """
    Convert a value to a mapping.

    Args:
        t (type): The type to cast mapping values to.
        sep (str): The separator to split the value by.
        from_file (bool): Whether to allow the value to be a readable from a file.
    """

    def __init__(
        self,
        t: type = None,
        mode: T.Literal["cast", "unpack"] = "cast",
        sep: str = ITER_SEP,
        from_file: bool = False,
    ):
        super().__init__(t, sep, from_file)
        self.mode = mode

    def parse(self, value) -> T.Mapping:
        if isinstance(value, str):
            if self.from_file and value.startswith(FROM_FILE_PREFIX):
                filepath = value[len(FROM_FILE_PREFIX) :]
                if os.path.isfile(filepath):
                    value = self.read_from_file(filepath)
                else:
                    raise FileNotFoundError(value)
            else:
                value = json.loads(value)
            if self.t is not None:
                if self.mode == "cast":
                    return self.t(value)
                elif self.mode == "unpack":
                    return self.t(**value)  # type: ignore
                else:
                    raise ValueError(f"Invalid mode: {self.mode}")
            return value
        raise TypeError(f"Cannot convert '{value}' to a mapping")

    def read_from_file(self, value: str) -> T.Mapping:  # type:ignore
        if value.endswith((".yaml", ".yml")):
            return yaml_load(value)  # type:ignore
        return json_load(value)  # type:ignore


class DatetimeParser(Parser):
    def parse(self, value) -> datetime.datetime:
        return dt.parse_datetime_str(value)


class DateParser(Parser):
    def parse(self, value) -> datetime.date:
        return dt.parse_datetime_str(value).date()


class SliceParser(Parser):
    t = slice

    def __init__(self, sep: str = SLICE_SEP):
        self.sep = sep

    def _get_nums(self, value: str):
        return [float(i) if i else None for i in value.split(self.sep)]

    def parse(self, value: str):
        if isinstance(value, self.t):
            return value
        nums = self._get_nums(value)
        if len(nums) not in (1, 2, 3):
            raise ValueError(
                f"{self.t.__name__} argument must be 1-3 values separated by '{self.sep}'"
            )
        return self.t(*nums)


class RangeParser(SliceParser):
    t = range

    def _get_nums(self, value: str):  # type:ignore
        return [int(i) for i in value.split(self.sep) if i]


class IntFloatParser(Parser):
    def parse(self, value: str) -> int | float:
        value = self.clean(value)
        if isinstance(value, (int, float)):
            return value
        if "e" in value or "." in value:
            return float(value)
        return int(value)


class BoolParser(Parser):
    def parse(self, value: str) -> bool:
        if isinstance(value, bool):
            return value
        if isinstance(value, int):
            return bool(value)
        if isinstance(value, str):
            value = value.lower()
            if value in ["1", "true"]:
                return True
            if value in ["0", "false"]:
                return False
            raise ValueError(value)
        raise TypeError(value)


__all__ = [
    "Cast",
    "Parser",
    "DateParser",
    "DatetimeParser",
    "IterableParser",
    "MappingParser",
    "RangeParser",
    "SliceParser",
    "BoolParser",
]



import enum
import inspect
import json
import typing as T
from functools import partial

from objinspect.typing import (
    get_literal_choices,
    is_direct_literal,
    is_generic_alias,
    is_iterable_type,
    is_mapping_type,
    is_union_type,
    type_args,
    type_origin,
)

from strto.constants import ITER_SEP
from strto.parsers import Parser


class StrToTypeParser:
    def __init__(
        self,
        parsers: dict[T.Any, Parser | T.Callable[[str], T.Any]] | None = None,
    ) -> None:
        self.parsers: dict[T.Any, Parser | T.Callable[[str], T.Any]] = parsers or {}

    def __len__(self):
        return len(self.parsers)

    def __getitem__(self, t: T.Any):
        return self.get(t)

    def add(self, t: T.Any, parser: Parser | T.Callable[[str], T.Any]):
        self.parsers[t] = parser

    def extend(self, parsers: dict[T.Any, Parser | T.Callable[[str], T.Any]]):
        self.parsers.update(parsers)

    def get(self, t: T.Any) -> Parser | T.Callable[[str], T.Any]:
        return self.parsers[t]

    def get_parse_func(self, t: T.Any) -> T.Callable[[str], T.Any]:
        return partial(self.parse, t=t)

    def parse(self, value: str, t: T.Any) -> T.Any:
        if parser := self.parsers.get(t, None):
            return parser(value)
        if is_generic_alias(t):
            return self._parse_alias(value, t)
        if is_union_type(t):
            return self._parse_union(value, t)
        if value is None:
            return None
        return self._parse_special(value, t)

    def _parse_alias(self, value: str, t: T.Any):
        base_t = type_origin(t)
        sub_t = type_args(t)

        if is_mapping_type(base_t):
            key_type, value_type = sub_t
            mapping_instance = base_t()
            for k, v in json.loads(value).items():
                mapping_instance[self.parse(k, key_type)] = self.parse(v, value_type)
            return mapping_instance

        elif is_iterable_type(base_t):
            item_t = sub_t[0]
            return base_t([self.parse(i.strip(), item_t) for i in value.split(ITER_SEP)])

        raise NotImplementedError

    def _parse_union(self, value: str, t: T.Any) -> T.Any:
        for t in type_args(t):
            try:
                return self.parse(value, t)
            except Exception:
                continue
        raise ValueError(f"Could not parse '{value}' as '{t}'")

    def _parse_special(self, value: str, t: T.Any) -> T.Any:
        """
        Parse enum or literal
        """
        if inspect.isclass(t):
            if issubclass(t, enum.Enum):
                return t[value]

        if is_direct_literal(t):
            choices = get_literal_choices(t)
            if not choices:
                raise ValueError("Empty literal")
            if value in choices:
                return value

            possible_types = [i for i in [int, str, bytes, bool] if i in {type(j) for j in choices}]
            for t in possible_types:
                try:
                    if t is bool:
                        parsed = value.lower() == "true"
                    else:
                        parsed = t(value)
                    if parsed in choices:
                        return parsed
                except ValueError:
                    continue
            else:
                raise ValueError(
                    f"'{value}' is not a valid choice for {t}. Valid choices are: {choices}"
                )
        raise NotImplementedError(t)


def get_parser(from_file: bool = True) -> StrToTypeParser:
    """
    Args:
        from_file (bool): Allow iterable types to be loaded from a file.
    """
    import datetime
    import decimal
    import fractions
    import pathlib
    from collections import ChainMap, Counter, OrderedDict, defaultdict, deque

    from strto.parsers import (
        BoolParser,
        Cast,
        DateParser,
        DatetimeParser,
        IntFloatParser,
        IterableParser,
        MappingParser,
        RangeParser,
        SliceParser,
    )

    DIRECTLY_CASTABLE_TYPES = [
        str,
        int,
        float,
        decimal.Decimal,
        fractions.Fraction,
        pathlib.Path,
        pathlib.PosixPath,
        pathlib.WindowsPath,
        pathlib.PureWindowsPath,
        pathlib.PurePosixPath,
        bytearray,
    ]

    ITERABLE_TYPES = [list, frozenset, deque, set, tuple]
    MAPPING_TYPES_CAST = [dict, OrderedDict, Counter]
    MAPPING_TYPES_UNPACK = []

    parser = StrToTypeParser()
    for t in DIRECTLY_CASTABLE_TYPES:
        parser.add(t, Cast(t))
    for t in MAPPING_TYPES_CAST:
        parser.add(t, MappingParser(t, from_file=from_file, mode="cast"))
    for t in MAPPING_TYPES_UNPACK:
        parser.add(t, MappingParser(t, from_file=from_file, mode="unpack"))
    for t in ITERABLE_TYPES:
        parser.add(t, IterableParser(t, from_file=from_file))

    int_float_parser = IntFloatParser()
    parser.extend(
        {
            bool: BoolParser(),
            range: RangeParser(),
            slice: SliceParser(),
            datetime.datetime: DatetimeParser(),
            datetime.date: DateParser(),
            int | float: int_float_parser,
            float | int: int_float_parser,
            T.Union[float, int]: int_float_parser,
            T.Union[int, float]: int_float_parser,
        }
    )
    return parser


__all__ = ["StrToTypeParser", "get_parser"]



import datetime
import enum
import fractions
import json
import pathlib
from collections import ChainMap, Counter, OrderedDict, defaultdict, deque
from fractions import Fraction

import pytest

from strto import StrToTypeParser, get_parser


@pytest.fixture
def parser() -> StrToTypeParser:
    return get_parser(from_file=True)


def test_simple_types(parser: StrToTypeParser):
    assert parser.parse("1.5", float) == 1.5
    assert parser.parse("5", int) == 5
    assert parser.parse("1/3", fractions.Fraction) == fractions.Fraction("1/3")
    assert parser.parse("./data/here.txt", pathlib.Path) == pathlib.Path("./data/here.txt")
    assert parser.parse('{"a":1,"b":2,"c":3}', dict) == {"a": 1, "b": 2, "c": 3}
    assert parser.parse("1,2,3,4,5", list) == ["1", "2", "3", "4", "5"]
    assert parser.parse(" 1 ,2,3   ,4   ,5", list) == ["1", "2", "3", "4", "5"]


def test_iter_alias_simple(parser: StrToTypeParser):
    assert parser.parse("1,2,3,4,5", list[int]) == [1, 2, 3, 4, 5]
    assert parser.parse("1,2,3,4,5", list[float]) == [1.0, 2.0, 3.0, 4.0, 5.0]
    assert parser.parse("1,2,3,4,5", list[str]) == ["1", "2", "3", "4", "5"]
    assert parser.parse("1/2,2/2,3/2,4/2,5/2", list[Fraction]) == [
        Fraction(1, 2),
        Fraction(2, 2),
        Fraction(3, 2),
        Fraction(4, 2),
        Fraction(5, 2),
    ]

    assert parser.parse("1,2,3,4,5", tuple[int]) == (1, 2, 3, 4, 5)
    assert parser.parse("1,2,3,4,5", tuple[float]) == (1.0, 2.0, 3.0, 4.0, 5.0)
    assert parser.parse("1,2,3,4,5", tuple[str]) == ("1", "2", "3", "4", "5")

    assert parser.parse("1,2,3,4,5", set[int]) == {1, 2, 3, 4, 5}
    assert parser.parse("1,2,3,4,5", set[float]) == {1.0, 2.0, 3.0, 4.0, 5.0}
    assert parser.parse("1,2,3,4,5", set[str]) == {"1", "2", "3", "4", "5"}

    assert parser.parse("1,2,3,4,5", frozenset[int]) == frozenset({1, 2, 3, 4, 5})
    assert parser.parse("1,2,3,4,5", frozenset[float]) == frozenset({1.0, 2.0, 3.0, 4.0, 5.0})
    assert parser.parse("1,2,3,4,5", frozenset[str]) == frozenset({"1", "2", "3", "4", "5"})


def test_mapping_alias(parser: StrToTypeParser):
    assert parser.parse('{"a":1,"b":2,"c":3}', dict[str, int]) == {"a": 1, "b": 2, "c": 3}
    assert parser.parse('{"a":1,"b":2,"c":3}', dict[str, float]) == {"a": 1.0, "b": 2.0, "c": 3.0}
    assert parser.parse('{"a":1,"b":2,"c":3}', dict[str, str]) == {"a": "1", "b": "2", "c": "3"}
    assert parser.parse('{"a":1,"b":2.5,"c":3.1}', dict[str, float | int]) == {
        "a": 1,
        "b": 2.5,
        "c": 3.1,
    }
    assert parser.parse('{"a":1,"b":2.5,"c":3.1}', dict[str, int | float]) == {
        "a": 1,
        "b": 2.5,
        "c": 3.1,
    }


def test_slice(parser: StrToTypeParser):
    assert parser.parse("0:5", slice) == slice(0, 5)
    assert parser.parse(":5", slice) == slice(None, 5)
    assert parser.parse("0:", slice) == slice(0, None)
    assert parser.parse("0:5:2", slice) == slice(0, 5, 2)
    assert parser.parse("::5", slice) == slice(None, None, 5)
    with pytest.raises(ValueError):
        parser.parse("0:5:2:4", slice)


def test_range(parser: StrToTypeParser):
    assert parser.parse("5", range) == range(5)
    assert parser.parse("5:6", range) == range(5, 6)
    assert parser.parse("0:5", range) == range(0, 5)
    assert parser.parse(":5", range) == range(0, 5)
    assert parser.parse("5:", range) == range(5)
    assert parser.parse("0:5:2", range) == range(0, 5, 2)
    with pytest.raises(ValueError):
        parser.parse("0:5:2:4", range)


def test_datetime(parser: StrToTypeParser):
    date = datetime.datetime(year=2022, day=19, month=7)
    assert parser.parse("2022.07.19", datetime.datetime) == date
    assert parser.parse("2022/07/19", datetime.datetime) == date
    assert parser.parse("19-7-2022", datetime.datetime) == date
    assert parser.parse("July 19th 2022", datetime.datetime) == date


def test_date(parser: StrToTypeParser):
    date = datetime.date(year=2022, day=19, month=7)
    assert parser.parse("2022.07.19", datetime.date) == date
    assert parser.parse("2022/07/19", datetime.date) == date
    assert parser.parse("19-7-2022", datetime.date) == date
    assert parser.parse("July 19th 2022", datetime.date) == date


def test_enum(parser: StrToTypeParser):
    class MyEnum(enum.Enum):
        A = 1
        B = 2
        C = 3

    assert parser.parse("A", MyEnum) == MyEnum.A
    with pytest.raises(KeyError):
        parser.parse("D", MyEnum)


def test_literal(parser: StrToTypeParser):
    from typing import Literal

    MyLiteral = Literal[1, 2, 3]
    assert parser.parse("1", MyLiteral) == 1
    assert parser.parse("False", Literal[True, False]) == False
    assert parser.parse(b"bytes", Literal[b"bytes", b"string"]) == b"bytes"

    with pytest.raises(ValueError):
        parser.parse("4", MyLiteral)
